# ---------------------------------------------------------------------------------------------------------
# THIS FILE CONTAINS THE FUNCTIONS TO PERFORM THE CHOW LIU ALGORITHM
# 1) build Chow-Liu tree with empirical covariance matrix (ChowLiu2)
# 2) ChowLiuFromSet: add one extra variable, constrain the edges to its children to exist and build the 
#    rest of the Chow Liu tree
# 3) convert index of matrix in row/column indices (returnCouple)
# ---------------------------------------------------------------------------------------------------------


ChowLiu2 <- function(Sigma){
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Build Chow-Liu tree from empirical covariance matrix
  # INPUT
  #   Sigma    : empirical covariance matrix (pxp matrix)
  # OUTPUT
  #   adjmat   : adjacency matrix describing conditional dependence structure
  # ---------------------------------------------------------------------------------------------------------
	df <- data.frame(Sigma)
	colnames(df) <- c(1:nrow(Sigma))
	graph <- chow.liu(df)
	adjmat <- amat(graph)
	return(adjmat)
}

ChowLiuFromSet <- function(S, set){
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Build Chow-Liu tree with one extra hidden variable
  # INPUT
  #   Sigma    : empirical covariance matrix (pxp matrix)
  #   set      : vector containing the indices of the children of the hidden variable
  # OUTPUT
  #   adjmat   : adjacency matrix describing conditional dependence structure ((p+1)x(p+1) matrix)
  # ---------------------------------------------------------------------------------------------------------
	p <- nrow(S)
	Sigma <- repmat(0,p+1,p+1)
	Sigma[1:p,1:p] <- S
	df <- data.frame(Sigma)
	colnames(df) <- c(1:(p+1))
	a <- length(set)
	vector <- rep(p+1,a)
	whitelist <- data.frame(cbind(set,vector))
	graph <- chow.liu(df, whitelist=whitelist)
	adjmat <- amat(graph)
	return(adjmat)
}

returnCouple <- function(ind, p){
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Convert linear matrix index into row/column indices
  # INPUT
  #   ind    : matrix index (integer in 1:p^2
  #   p      : size of matrix (integer)
  # OUTPUT
  #   couple : vector of size 2 containing the row and column indices to get the ind_th value of a matrix of
  #            size p
  # ---------------------------------------------------------------------------------------------------------
  
  if(ind%%p==0){
    i <- ind%/%p
    j <- p
  } else{
    i <- ind%/%p+1
    j <- ind%%p
  }
  couple <- c(i,j)
  return(couple)
}

