# ---------------------------------------------------------------------------------------------------------
# THIS FILE CONTAINS THE FUNCTIONS TO PERFORM THE HIERARCHICAL CLUSTERING PROCEDURE TO INITIALIZE EM
# 1) Build Chow-Liu graph with constrained cliques (GraphFromSet)
# 2) Extract the triplets of nodes in which at least one edge exists (getTriplets)
# 3) Order the possible groups to form in decreasing order of the BIC gain and perform 
#    the best merge action (orderLikelihoods) 
# 4) Compute the likelihood of merging together a group of nodes (computeLikelihood)
# 5) Compute the likelihood of the data under a given adjacency matrix and constrained cliques (likelihood)
# 6) Extract the best first group of three nodes to group together based on likelihood gain (initTriplet) 
# 7) Perform the hierarcical classification (hClassif)
# 8) Cut the hierarchy at the level that maximizes the BIC and extract the resulting clusters
#    (getInitCliques)
# ---------------------------------------------------------------------------------------------------------


GraphFromSet <- function(S, set) {
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Build graph with a constrained clique and the rest of the graph constructed with Chow Liu algorithm
  # INPUT
  #   Sigma    : empirical covariance matrix (pxp matrix)
  #   set      : vector containing the indices of the nodes in the clique
  # OUTPUT
  #   adjmat   : adjacency matrix describing conditional dependence structure (pxp matrix)
  #   K        : maximum likelihood estimator of precision matrix
  #   likelihood : likelihood of data under the constructed model
  # ---------------------------------------------------------------------------------------------------------
	k <- 1
  p <- nrow(S)
	Sigma <- repmat(0, p + k, p + k)
	Sigma[1:p, 1:p] <- S
	df <- data.frame(Sigma)
	colnames(df) <- c(1:(p + k))
	a <- length(set)
	vector <- rep(p + 1, a)
	whitelist <- data.frame(cbind(set, vector))
	graph <- chow.liu(df, whitelist = whitelist)
	adjmat <- amat(graph)
	E <- which(adjmat[1:p, 1:p] != 0, arr.ind = T)
	degrees <- rowSums(adjmat)[1:p]
	K.fill <- matrix(0, p, p)
	for (c in 1:nrow(E)) {
		i <- E[c, 1]
		j <- E[c, 2]
		K.fill[c(i, j), c(i, j)] <- K.fill[c(i, j), c(i, j)] + solve(S[c(i, j), c(i, j)])
	}
	K <- K.fill/2
	diag(K) <- diag(K) - (degrees - 1)/diag(S)
	K[set, set] <- K[set, set] + pinv(S[set, set])
	likelihood <- log(det(K)) - matrix.trace(K %*% S)
	adjmat_marg <- (K != 0) * 1
	diag(adjmat_marg) <- 0
	return(structure(list(adjmat = adjmat_marg, K = K, likelihood = likelihood)))
}

getTriplets <- function(adjmat, clique.idx = NULL) {
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Extract triplets of nodes among which there exists at least one edge
  # INPUT
  #   adjmat     : adjacency matrix describing the edges (pxp matrix)
  #   clique.idx : vector containing the indices of the nodes that already belong to an existing clique
  # OUTPUT
  #   good_triplets   : matrix of size 3xk containing the triplets among which there is at least one edge
  #                     good_triplets[,i] is a vector of size 3 containing the indices of the nodes
  #                     in the i_th triplet (the order does not mean anything)
  # ---------------------------------------------------------------------------------------------------------
	p <- nrow(adjmat)
	E <- which(adjmat != 0, arr.ind = T)
	degrees <- rowSums(adjmat)
	triplets <- combn(1:p, 3)
	edge_triplets <- sapply(1:ncol(triplets), function(k) length(intersect(triplets[, k], clique.idx)) == 0 && sum(adjmat[triplets[, 
		k], triplets[, k]]) > 0)
	good_triplets <- triplets[, which(edge_triplets)]
	return(good_triplets)
}


orderLikelihoods <- function(S, cliquelist, n, merge, mergelist) {
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Order the possible groups to form in decreasing order of the BIC gain and perform 
  #   the best merge action  
  # INPUT
  #   S          : empirical covariance matrix (pxp matrix)
  #   cliquelist : list containing the existing clusters of the hierarchy: cliquelist[[i]] is a vector
  #                containing the indices of the nodes in the ith cluster
  #   n          : sample size (for the BIC)
  #   merge      : matrix of size mx2 where m is the current iteration of the hierarchical clustering.
  #                see the documentation of hclust for more details
  #   mergelist  : list of same size as cliquelist. mergelist[[i]] contains the stage at which 
  #                cliquelist[[i]] was formed. For example if the initial triplet is at position i 
  #                in cliquelist then mergelist[[i]]=1
  # OUTPUT
  #   cliquelist   : updated list of cliques after performing the best merge step
  #   likelihood   : value of BIC after performing the merge
  #   labels       : vectors of size p containing the labels of the clustering at this stage of HC
  #   mergelist    : updated value of the mergelist list
  #   merge        : updated value of the merge matrix
  #   flag         : boolean: TRUE if the BIC has increased after the merge, FALSE otherwise
  # ---------------------------------------------------------------------------------------------------------
	k <- length(cliquelist)
	p <- nrow(S)
	Sigma <- repmat(0, p + k, p + k)
	Sigma[1:p, 1:p] <- S
	df <- data.frame(Sigma)
	colnames(df) <- c(1:(p + k))
	clique.idx <- unlist(cliquelist)
	nodelist <- setdiff(1:p, clique.idx)
	set <- cliquelist[[1]]
	vector <- rep(p + 1, length(set))
	whitelist <- cbind(set, vector)

	if (k > 2) {
		for (i in 2:k) {
			vector <- rep(p + i, length(cliquelist[[i]]))
			whitelist <- rbind(whitelist, cbind(cliquelist[[i]], vector))

		}
	}
	whitelist <- data.frame(whitelist)
	graph <- chow.liu(df, whitelist = whitelist)
	adjmat <- amat(graph)[1:p, 1:p]
	K <- sapply(cliquelist, function(x) length(x) * (length(x) - 1))
	K <- sum(K)
	l <- likelihood(S, adjmat, cliquelist)$likelihood - 0.25 * log(n)/n * (sum(adjmat[1:p, 1:p])-p)
	tripletlist <- getTriplets(adjmat, clique.idx)
	testlist <- list()
	if (k >= 2) {
		couples1 <- combn(1:k, 2)
		for (i in 1:ncol(couples1)) {
			c <- couples1[, i]
			idx1 <- c[1]
			idx2 <- c[2]
			clique1 <- cliquelist[[idx1]]
			clique2 <- cliquelist[[idx2]]
			testlist[[i]] <- list(clique1, clique2)
		}
	}

	offset1 <- length(testlist)
	if (length(nodelist) > 0) {
		couples2 <- expand.grid(1:k, 1:length(nodelist))
		for (i in 1:nrow(couples2)) {
			c <- as.matrix(couples2[i, ])
			idx1 <- c[1]
			idx2 <- c[2]
			clique1 <- cliquelist[[idx1]]
			clique2 <- nodelist[idx2]
			testlist[[offset1 + i]] <- list(clique1, clique2)
		}
	}

	offset2 <- length(testlist)
	if (!is.null(ncol(tripletlist)) && !ncol(tripletlist)==0){
		for (i in 1:ncol(tripletlist)) {
			testlist[[offset2 + i]] <- tripletlist[, i]
		}
	}

	testlength <- length(testlist)
	likelihoods <- rep(0, testlength)
	cliquelist2 <- list()
	cases <- list()
	if (testlength >= 1) {
		for (i in 1:testlength) {
			if (i%%100 == 0) {
				print(i)
			}
			res <- computeLikelihood(S, cliquelist, testlist[[i]], n)
			K <- res$K
			likelihoods[i] <- res$likelihood
			cliquelist2[[i]] <- res$cliquelist
			cases[[i]] <- res$case
		}
		lik <- max(likelihoods, na.rm = TRUE)
			idx <- which(likelihoods == lik)
			idx <- idx[1]
			finalclique <- cliquelist2[[idx]]
			k <- length(finalclique)
			labels <- rep(0, p)
			case <- cases[[idx]]
			for (i in 1:k) {
				labels[finalclique[[i]]] <- i
			}
			nodeset <- setdiff(1:p, unlist(finalclique))
			if(length(nodeset>=1)){ 
				for (i in 1:length(nodeset)) {
					labels[nodeset[i]] <- (k + i)
				}
			}

			flag <- TRUE
			set <- testlist[[idx]]
			if (case == 1) {
				clust <- max(unlist(mergelist))+1
				merge <- rbind(merge, c(-set[1], -set[2]))
				merge <- rbind(merge, c(-set[3], clust))
				mergelist[[length(mergelist) + 1]] <- clust + 1
			} else if (case == 3) {
				set.indices <- couples1[, idx]
				idx1 <- set.indices[1]
				idx2 <- set.indices[2]
				set1 <- cliquelist[[idx1]]
				set2 <- cliquelist[[idx2]]
				merge <- rbind(merge, c(mergelist[[idx1]], mergelist[[idx2]]))

				mergelist.tmp <- list()
				it <- 1
				for (i in 1:length(mergelist)) {
					if (length(intersect(cliquelist[[i]], set1)) == length(set1)) {
						mergelist.tmp[[it]] <- max(unlist(mergelist)) + 1
						it <- it + 1
					} else if (length(intersect(cliquelist[[i]], set2)) == 0) {
						mergelist.tmp[[it]] <- mergelist[[i]]
						it <- it + 1
					}
				}
				mergelist <- mergelist.tmp


			} else {
				set.indices <- couples2[idx - offset1,]
				idx1 <- set.indices[1]
				idx2 <- nodelist[unlist(set.indices[2])]
				set1 <- cliquelist[[unlist(idx1)]]
				merge <- rbind(merge, c(mergelist[[unlist(idx1)]], -idx2))
				mergelist.tmp <- mergelist
				for (i in 1:length(mergelist)) {
					if (length(intersect(cliquelist[[i]], set1)) == length(set1)) {
						mergelist.tmp[[i]] <- max(unlist(mergelist)) + 1
					}
				}
				mergelist <- mergelist.tmp
			}

		} 	else {
		finalclique <- cliquelist
		likelihood <- l
		labels <- labels
		flag <- FALSE
	}

	return(structure(list(likelihood = lik, cliquelist = finalclique, labels = labels, flag = flag, merge = merge, mergelist = mergelist)))
}

computeLikelihood <- function(S, cliquelist, set, n, merge) {
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Compute the BIC of the data after creating a new cluster
  # INPUT
  #   S          : empirical covariance matrix (pxp matrix)
  #   cliquelist : list containing the existing clusters of the hierarchy: cliquelist[[i]] is a vector
  #                containing the indices of the nodes in the ith cluster
  #   n          : sample size (for the BIC)
  #   merge      : matrix of size mx2 where m is the current iteration of the hierarchical clustering.
  #                see the documentation of hclust for more details
  #   set        : vector containing the indices that should be merged. It can be 
  #                1) a new triplet -> case = 1 2) an element of cliquelist + a new node -> case = 2
  #                3) 2 elements of cliquelist -> case = 3
  # OUTPUT
  #   cliquelist   : updated list of cliques after performing the best merge step
  #   likelihood   : value of BIC after performing the merge
  #   K            : maximum likelihood estimator of precision matrix with corresponding support
  #   case         : integer (1,2 or 3) giving the type of the cluster formed as described in set INPUT
  # ---------------------------------------------------------------------------------------------------------
	k <- length(cliquelist)
	if (length(set) == 3) {
		case <- 1
		cliquelist.tmp <- cliquelist
		cliquelist.tmp[[k + 1]] <- set
	} else if (length(set[[2]]) == 1) {
		case <- 2
		cliquelist.tmp <- cliquelist
		for (i in 1:k) {
			if (length(intersect(cliquelist[[i]], set[[1]])) == length(set[[1]])) {
				cliquelist.tmp[[i]] <- c(set[[1]], set[[2]])
			}
		}
	} else {
		case <- 3
		cliquelist.tmp <- list()
		it <- 1
		for (i in 1:k) {
			if (length(intersect(cliquelist[[i]], set[[1]])) == length(set[[1]])) {
				cliquelist.tmp[[it]] <- c(set[[1]], set[[2]])
				it <- it + 1
			} else if (length(intersect(cliquelist[[i]], set[[2]])) == 0) {
				cliquelist.tmp[[it]] <- cliquelist[[i]]
				it <- it + 1
			}
		}
	}

	k <- length(cliquelist.tmp)
	p <- nrow(S)
	Sigma <- repmat(0, p + k, p + k)
	Sigma[1:p, 1:p] <- S
	df <- data.frame(Sigma)
	colnames(df) <- c(1:(p + k))
	clique.idx <- unlist(cliquelist.tmp)
	nodelist <- setdiff(1:p, clique.idx)
	set1 <- cliquelist.tmp[[1]]
	vector <- rep(p + 1, length(set1))
	whitelist <- cbind(set1, vector)

	if (k >= 2) {
		for (i in 2:k) {
			vector <- rep(p + i, length(cliquelist.tmp[[i]]))
			whitelist <- rbind(whitelist, cbind(cliquelist.tmp[[i]], vector))

		}
	}
	whitelist <- data.frame(whitelist)
	graph <- chow.liu(df, whitelist = whitelist)
	adjmat <- amat(graph)
	res <- likelihood(S, adjmat, cliquelist.tmp)
	K <- res$K
	likelihood <- res$likelihood - 0.25 * (sum(K != 0) - p) * log(n)/n
	return(structure(list(likelihood = likelihood, cliquelist = cliquelist.tmp, K = K, case = case)))
}


likelihood <- function(S, adjmat, cliquelist) {
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Compute the BIC of the data under constrained adjacency matrix and cliques
  # INPUT
  #   S          : empirical covariance matrix (pxp matrix)
  #   cliquelist : list containing the existing clusters of the hierarchy: cliquelist[[i]] is a vector
  #                containing the indices of the nodes in the ith cluster
  #   adjmat     : adjacency matrix (pxp matrix)
  # OUTPUT
  #   likelihood   : value of BIC 
  #   K            : maximum likelihood estimator of precision matrix with corresponding support
  # ---------------------------------------------------------------------------------------------------------
	p <- nrow(S)
	E <- which(adjmat[1:p, 1:p] != 0, arr.ind = T)
	degrees <- rowSums(adjmat)[1:p]
	K.fill <- matrix(0, p, p)
	for (c in 1:nrow(E)) {
		i <- E[c, 1]
		j <- E[c, 2]
		K.fill[c(i, j), c(i, j)] <- K.fill[c(i, j), c(i, j)] + solve(S[c(i, j), c(i, j)])
	}
	K <- K.fill/2
	diag(K) <- diag(K) - (degrees - 1)/diag(S)
	for (i in 1:length(cliquelist)) {
		if (length(cliquelist[[i]]) > 1) {
			K[cliquelist[[i]], cliquelist[[i]]] <- K[cliquelist[[i]], cliquelist[[i]]] + pinv(S[cliquelist[[i]], cliquelist[[i]]])
		}
	}
	likelihood <- log(det(K)) - matrix.trace(K %*% S)
	return(structure(list(likelihood = likelihood, K = K)))
}

initTriplet <- function(S, n) {
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Extract the initial triplet with highest for which the merge results in the highest BIC gain 
  # INPUT
  #   S          : empirical covariance matrix (pxp matrix)
  #   n          : sample size (for the BIC)
  # OUTPUT
  #   likelihood : value of BIC after performing the merge
  #   merge      : matrix of size mx2 where m is the current iteration of the hierarchical clustering.
  #                see the documentation of hclust for more details
  #   K            : maximum likelihood estimator of precision matrix with corresponding support
  #   mergelist  : list of same size as cliquelist. mergelist[[i]] contains the stage at which 
  #                cliquelist[[i]] was formed. For example if the initial triplet is at position i 
  #                in cliquelist then mergelist[[i]]=1
  #   labels       : vectors of size p containing the labels of the clustering at this stage of HC
  #   flag         : boolean: TRUE if the BIC has increased after the merge, FALSE otherwise
  # ---------------------------------------------------------------------------------------------------------
	p <- nrow(S)
	Tree <- ChowLiu2(S)
	E <- which(Tree != 0, arr.ind = T)
	degrees <- rowSums(Tree)
	K.fill <- matrix(0, p, p)
	for (c in 1:nrow(E)) {
		i <- E[c, 1]
		j <- E[c, 2]
		K.fill[c(i, j), c(i, j)] <- K.fill[c(i, j), c(i, j)] + solve(S[c(i, j), c(i, j)])
	}
	K_CL <- K.fill/2
	diag(K_CL) <- diag(K_CL) - (degrees - 1)/diag(S)
	l <- log(det(K_CL)) - matrix.trace(K_CL %*% S) - 0.25 * (sum(K_CL != 0) -p)* log(n)/n
	triplets <- getTriplets(Tree)
	likelihoods <- rep(0, ncol(triplets))
	for (k in 1:ncol(triplets)) {
		if (k%%100 == 0) {
			print(k)
		}
		t <- triplets[, k]
		res <- GraphFromSet(S, set = t)
		K <- res$K
		likelihoods[k] <- res$likelihood - 0.25 * (sum(K != 0) -p)* log(n)/n
	}
	lik <- max(likelihoods)
	best.idx <- which(likelihoods == lik)
	best.triplet <- triplets[, best.idx]
	labels <- rep(0, p)
	if (lik > l) {
		flag <- TRUE
		for (i in best.triplet) {
			labels[i] <- 1
		}
		nodeset <- setdiff(1:p, best.triplet)
		for (i in 1:length(nodeset)) {
			labels[nodeset[i]] <- (1 + i)
		}

	} else {
		flag <- FALSE
	}
	merge <- rbind(c(-best.triplet[1], -best.triplet[2]), c(-best.triplet[3], 1))
	mergelist <- list(2)
	return(structure(list(likelihood = max(likelihoods), triplet = best.triplet, flag = flag, labels = labels, merge = merge, 
		mergelist = mergelist)))
}

hclassif <- function(S, n) {
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Extract the initial triplet with highest for which the merge results in the highest BIC gain 
  # INPUT
  #   S          : empirical covariance matrix (pxp matrix)
  #   n          : sample size (for the BIC)
  # OUTPUT
  #   cliquelist : list containing the clusters of the hierarchy at height where BIC is the highest.
  #                cliquelist[[i]] is a vector containing the indices of the nodes in the ith cluster. 
  # ---------------------------------------------------------------------------------------------------------
  
	p <- nrow(S)
	res <- initTriplet(S, n)
	cliquelist <- list(res$triplet)
	labelslist <- list()
	likelihoods <- list()
	likelihoods[[1]] <- res$likelihood
	labelslist[[1]] <- res$labels
	merge <- res$merge
	mergelist <- res$mergelist
	iter <- 1
	flag <- length(cliquelist) == 2 && length(cliquelist[[1]]) + length(cliquelist[[2]]) == p
	while (!flag) {
	#while(iter<10) {
		res <- orderLikelihoods(S, cliquelist, n, merge, mergelist)
		merge <- res$merge
		mergelist <- res$mergelist
		cliquelist <- res$cliquelist
		flag <- length(res$cliquelist) == 2 && length(res$cliquelist[[1]]) + length(res$cliquelist[[2]]) == p
		likelihoods[[iter + 1]] <- res$likelihood
		labelslist[[iter + 1]] <- res$labels
		iter <- iter + 1
	}
	return(structure(list(likelihoods = likelihoods, labelslist = labelslist, flag = flag, merge = merge, mergelist=mergelist)))
}


getInitCliques <- function(S,n){
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Cut the tree where the BIC is the highest
  # INPUT
  #   S          : empirical covariance matrix (pxp matrix)
  #   n          : sample size (for the BIC)
  # OUTPUT
  #   res : result of hierarchical classif
  #   cliquelist  : list of clusters at highest value of BIC
  # ---------------------------------------------------------------------------------------------------------
  res <- hclassif(S,n)
  likelihoods <- unlist(res$likelihoods)
  idx <- which(likelihoods==max(likelihoods))
  labelslist <- res$labelslist
  labels <- labelslist[[1+idx]]
  labels <- labelslist[[1+idx]]
  labelmax <- max(labels)
  num <- lapply(1:labelmax, function(i) length(which(labels == i)))
  num <- unlist(num)
  num_clusts <- max(which(num>1))
  cliquelist <- lapply(1:num_clusts, function(k) which(labels == k))
  return(structure(list(res=res, cliquelist=cliquelist)))
}







