# ---------------------------------------------------------------------------------------------------------
# THIS FILE CONTAINS THE FUNCTIONS TO PERFORM THE MEINSHAUSEN & BUHLMANN APPROX
# 1) find the regularization parameter lambda that drives all the coefficients to 0 (get.lambda.l1)
# 2) Perform M&B approx for a range of lambdas (MeBuTest)
# ---------------------------------------------------------------------------------------------------------

# Meinshausen & Bühlmann ----------------------------------------------------------------
get.lambda.l1 <- function(S) {
  r.max <- max(0,max(abs(S-diag(diag(S)))))
  return(r.max)
}


MeBuTest <- function(S, ground_truth, nlambda=10, log.rho.min=-5){
  # ---------------------------------------------------------------
  # FUNCTION
  #   Build glasso graph from Covariance matrix of multivariate Gaussian with Meishausen & Bühlmann
  #   approximation. Returns result with best accuracy with respect to specified ground truth for
  #   a set of regularization parameters
  # INPUT
  #   Sigma: Covariance matrix
  # OUTPUT
  #   adjmat: Adjacency matrix describing conditional dependencies discovered with glasso
  # ---------------------------------------------------------------
  rho.max <- get.lambda.l1(S)
  log.rho.max <- log(rho.max)
  log.rho <- seq(log.rho.min, log.rho.max, length.out=nlambda)
  Ks <- lapply(exp(log.rho), function(x) glasso(S, x, trace=TRUE, approx=TRUE))
  wis <- lapply(Ks, function(K) K$wi)
  GraphsEstimates <- lapply(wis, function(wi) graphFromPCor(wi))
  accuracy <- sapply(1:length(log.rho), function(x) sum(ground_truth==GraphsEstimates[[x]]))
  idx <- sort.list(accuracy, decreasing=TRUE)[1]
  adjmat <- GraphsEstimates[[idx]]
  return(adjmat)
}

