# ---------------------------------------------------------------------------------------------------------
# THIS FILE CONTAINS THE FUNCTIONS TO PERFORM THE TREE AGREGATION PROCEDURE
# 1) Compute partition function with minors (computeZ)
# 2) Compute the Z(kl) with minors (computeZkl)
# 3) Compute the Alpha matrix containing the posterior edge probabilities (computeAlpha)
# 4) Compute the K matrix necessary in the EM procedure (computeKtilde)
# 5) Build a graph with this procedure (constructTreeAgregation)
# ---------------------------------------------------------------------------------------------------------

computeZ <- function(W, nb.runs=10){
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Compute partition function of posterior edges probabilities for a tree using laplacian minors
  #   Z is computed by taking the average of several minors
  # INPUT
  #   W        : matrix containing edge weights (pxp matrix)
  #   nb.runs  : number of iterations in the averaging (integer, default is 10)
  # OUTPUT
  #   Z    : partition function
  # ---------------------------------------------------------------------------------------------------------
  Delta <- -W
  diag(Delta) <- rowSums(W)
  p <- nrow(Delta)
  minor <- 0
  for(i in 1:nb.runs){
    k <- sample(1:p,1)
    l <- sample(setdiff(1:p,k),1)
    columns <- setdiff(1:p,l)
    rows <- setdiff(1:p,k)
    Deltaprime <- Delta[rows, columns]
    minor <- minor + abs(det(Deltaprime))
  }
  Z <- minor/nb.runs
  return(Z)
}

computeZkl <- function(W, k, l){
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Compute the Zkl for posterior edges probabilities for a tree using laplacian minors
  # INPUT
  #   W        : matrix containing edge weights (pxp matrix)
  #   k        : row index (integer in 1:p)
  #   l        : column index (integer in 1:p)
  # OUTPUT
  #   Z_kl    : Z_kl
  # ---------------------------------------------------------------------------------------------------------
  p <- nrow(W)
  W_kl <- W
  W_kl[k,l] <- 0
  W_kl[l,k] <- 0
  Z_kl <- computeZ(W_kl)
  return(Z_kl)
}

computeAlpha <- function(S){
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Compute the matrix of posterior edge probabilities
  # INPUT
  #   S        : empirical covariance matrix (pxp matrix)
  # OUTPUT
  #   Alpha    : matrix of posterior edge probabilities (pxp matrix)
  # ---------------------------------------------------------------------------------------------------------

  D <- diag(1/sqrt(diag(S)))
  Cor <- D %*% S %*% D
  p <- nrow(S)
  diag(Cor) <- 0
  InfoMat <- -0.5 * log(1 - Cor^2) * lower.tri(Cor)
  W <- exp(InfoMat+t(InfoMat))
  diag(W) <- 0
  Z <- computeZ(W)
  Alpha <- sapply(1:p, function(k) sapply(1:p, function(l) 1-computeZkl(W, k, l)/Z))*lower.tri(W)
  A <- order(Alpha, decreasing = TRUE)
  A <- A[1:(p * (p - 1)/2)]
  ordered_couples <- lapply(A, function(x) returnCouple(x, p))
  return(structure(list(Alpha      = Alpha + t(Alpha),
                        couples    = ordered_couples)))
}

computeKtilde <- function(Alpha,S){
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Compute the Ktilde matrix necessary in the EM procedure (expectation of precision matrix K under the
  #   posterior edge distribution)
  # INPUT
  #   S        : empirical covariance matrix (pxp matrix)
  #   Alpha    : matrix of posterior edge probabilities (pxp matrix)
  # OUTPUT
  #   Ktilde    : expectation of precision matrix K under the posterior edge distribution (pxp matrix)
  # ---------------------------------------------------------------------------------------------------------
  p <- nrow(S)
  E <- combn(1:p,2)
  K.fill <- matrix(0,p,p)
  for (c in 1:ncol(E)) {
    i <- E[1, c]
    j <- E[2, c]
    K.fill[c(i, j), c(i, j)] <- K.fill[c(i, j), c(i, j)] + Alpha[i,j]*solve(S[c(i, j), c(i, j)])
  }
  Ktilde <- K.fill/2
  degrees <- rowSums(Alpha)
  diag(Ktilde) <- diag(Ktilde) - (degrees - 1)/diag(S)
  return(Ktilde)
}

constructTreeAgregation <- function(S,m){
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Graph construction using Bayesian agregation of trees
  # INPUT
  #   S        : empirical covariance matrix (pxp matrix)
  #   m        : number of edges wanted in the graph (integer in 1:p(p-1))
  # OUTPUT
  #   adjmat    : adjacency matrix describing the conditional independence structure (pxp matrix)
  # ---------------------------------------------------------------------------------------------------------

  p <- nrow(S)
  if(is.null(m)){
    m=p-1
  }
  res <- computeAlpha(S)
  ordered_couples <- res$couples
  adjmat <- matrix(0,p,p)
  for(i in 1:m){
    k <- ordered_couples[[i]][1]
    l <- ordered_couples[[i]][2]
    adjmat[k,l] <- 1
    adjmat[l,k] <- 1
  }
  return(adjmat)
}

constructGraphFromAlpha <- function(Alpha, m, ordered_couples){
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Graph construction using Bayesian agregation of trees
  # INPUT
  #   Alpha    : matrix of posterior edge probabilities (pxp matrix)
  #   m        : number of edges wanted in the graph (integer in 1:p(p-1))
  #   ordered_couples : list of edges in the decreasing order of the posterior probabilities
  #   list of size p(p-1)/2. The k_th element  of the list noted e is a vector of size 2 such that
  #   i = e[1] and j=e[2] are the indices of the nodes that have the k_th highest posterior probability
  #   of being connected by an edge
  # OUTPUT
  #   adjmat    : adjacency matrix describing the conditional independence structure (pxp matrix)
  # ---------------------------------------------------------------------------------------------------------

  p <- nrow(Alpha)
  if(is.null(m)){
    m=p-1
  }
  adjmat <- matrix(0,p,p)
  for(i in 1:m){
    k <- ordered_couples[[i]][1]
    l <- ordered_couples[[i]][2]
    adjmat[k,l] <- 1
    adjmat[l,k] <- 1
  }
  return(adjmat)
}



computeA <- function(Kh, alphah, Sigma){
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Compute the A coefficients of the M-step necessary to compute the K_ij, i€O and j€H
  # INPUT
  #   K        : K_OH (pxr matrix) p number of observed variables r number of hidden variables
  #   alpha    : matrix of posterior edge probabilities (pxp matrix)
  #   S        : Empirical covariance matrix of observed variables (pxp matrix)
  # OUTPUT
  #   A        : pxr matrix containing the coefficients A_ij of the Mstep. A[i,j]=A_ij
  # ---------------------------------------------------------------------------------------------------------
  Sh <- Sigma[1:p,p+1]
  if(is.null(ncol(Kh))){
    p <- length(Kh)
    r <- 1
  } else{
    p <- nrow(Kh)
    r <- ncol(Kh)
  }
  sgn <- matrix(0.5, p,p)
  diag(sgn) <- -1
  A <- 0.5*(Sh%*%Kh)*alphah + Sh*Kh*alphah
  # A <- (Sh%*%Kh)*(0.5*alphah+1)
  if(r>1){
    B <- K*alpha
    C <- lapply(1:p, function(i) B*S[i,])
    C2 <- lapply(1:p, function(k) sapply(1:r, function(j) sum(C[[k]][setdiff(1:p,k), setdiff(1:r,j)])))
    C <- t(do.call(abind, c(C2, along = 2)))
    A <- A+0.5*C
  }
  return(A)
}

computeF <- function(K, S, p){
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Compute the log(f_ih), i€O, h€H coefficients of the posterior edges probabilities
  # INPUT
  #   K        : (p+rxp+r matrix) p number of observed variables r number of hidden variables
  #   n        : number of samples
  #   S        : Empirical covariance matrix of observed variables (pxp matrix)
  # OUTPUT
  #   f        : pxr matrix containing the coefficients log(f_ih). f[i,h]=log(f_ih)
  # ---------------------------------------------------------------------------------------------------------

  if(is.null(ncol(K))){
    r <- length(K)-p
  } else{
    r <- ncol(K)-p
  }
  f <- (K[1:p,(p+1):(p+r)]%*%solve(K[(p+1):(p+r),(p+1):(p+r)]))*t(t(K[1:p,(p+1):(p+r)])%*%S)
  f <- f
  return(f)
}

optimDiag <- function(i,k_ii, K, Sigma, alpha){
  p <- nrow(K)
  k <- K[i,setdiff(1:p,i)]
  a <- alpha[i,setdiff(1:p,i)]
  return(1/k_ii-Sigma[i,i]+sum(a*k^2/(k_ii*diag(K)[setdiff(1:p,i)]-k^2)))
}


dichotomie <- function(a, b, f, epsilon){
  # ---------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   find the zero of monotonous function f by binary search.
  # INPUT
  #   a        : minimum argument of f
  #   b        : maximum argument of f
  #   epsilon  : tolerance (length of last interval)
  # OUTPUT
  #   c        : approximate zero of f with epsilon tolerance
  # ---------------------------------------------------------------------------------------------------------

  min <- a
  max <- b
  sgn <- sign(f(max)-f(min))
  c <- (max+min)/2
  while(abs(f(c))>1e-5){
    c <- (max+min)/2
    if(sgn*f(c)>0){
      max <- c
    } else{
      min <- c
    }
  }
  return(c)
}

getmin <- function(f,max){
  vals=f(seq(0,1,length.out=1000))
  mini=min(intersect(which(abs(vals)<Inf), which(sign(vals)+sign(f(max))==0)))*1e-3
}



