# ---------------------------------------------------------------------------------------------------------------------
# THIS FILE CONTAINS THE FUNCTIONS TO INFER A GAUSSIAN GRAPHICAL MODEL WITH DIFFERENT METHODS
# 1) Infer graphical model from empirical covariance matrix with several possible algorithms (inferAdjMat)
# ---------------------------------------------------------------------------------------------------------------------

inferAdjMat <- function(S, algo_type=c("MB", "CL", "RG"), XS=NULL, ground_truth=NULL, n=200){
  # -------------------------------------------------------------------------------------------------------------------
  # FUNCTION
  #   Infer graphical model from empirical covariance matrix with several possible algorithms
  # INPUT
  #   S            : empirical statistical distance matrix (pxp matrix)
  #   algo_type    : algorithm name among "MB"(Meinshausen & Bühlmann), "CL" (Chow-Liu) and "RG" (Recursive Grouping)
  #   XS           : data matrix (nxp matrix)
  #   ground_truth : true adjacency matrix if known to choose regularization parameter
  #   n            : sample size (integer)
  # OUTPUT
  #   adjmat   : adjacency matrix describing conditional dependencies ((p+k)x(p+k) matrix),
  #              k number of hidden variables
  #   edgeDist : estimator of statistical distance matrix ((p+k)x(p+k) matrix)
  # -------------------------------------------------------------------------------------------------------------------
  switch(algo_type,
         MB={
           adjmat <- MeBuTest(S, ground_truth=ground_truth)
         },
         CL={
           adjmat <- ChowLiu2(S)
         },
         RG={
           D <- diag(1/sqrt(diag(S)))
           Cor <- D%*%S%*%D
           p <- nrow(S)
           tiny  <-  1e-10
           Dist <- -log(abs(Cor)+tiny)
           Dist <- (Dist>0)*Dist
           diag(Dist) <- 0
           adjmat <- RG(Dist, n)$adjmat
         },
         stop("adjacency matrix estimated !")
         )

  ncond           <- nrow(S)
  adjmat_cond     <- adjmat[1:ncond, 1:ncond]
  ncomp           <- nrow(adjmat)

  if(ncomp>ncond){
    adjmat_hid <- adjmat[setdiff(1:ncomp,1:ncond), setdiff(1:ncomp,1:ncond)]
    if(length(setdiff(1:ncomp,1:ncond))>1){
      diag(adjmat_hid) <- 10
    }else{
      adjmat_hid <- 10
    }
    adjmat_inter      <- adjmat[1:ncond, setdiff(1:ncomp,1:ncond)]
    adjmat_marg       <- adjmat_cond+diag(10,ncond)+adjmat_inter%*%solve(adjmat_hid)%*%t(adjmat_inter)
    adjmat_marg       <- (adjmat_marg>0)*1
    diag(adjmat_marg) <- 0
  } else{
    adjmat_marg      <- adjmat_cond
  }
  return(structure(list(adjmat_cond    = adjmat_cond,
                        adjmat_marg    = adjmat_marg)))
}



