# ---------------------------------------------------------------------------------------------------------------------
# THIS FILE CONTAINS THE FUNCTIONS TO COMPUTE STATISTICS TO EVALUATE THE RESULT OF GRAPHICAL MODEL INFERENCE
# 1) Compute the False Discovery Rate (FDR) 
# 2) Compute the power (power)
# 3) Compute the Kullback-Leibler divergence (KLdiv)
# ---------------------------------------------------------------------------------------------------------------------

FPR <- function(adjmat0, adjmat1){
  # ---------------------------------------------------------------
  # FUNCTION
  #   Compute False Descovery Rate 
  # INPUT
  #   adjmat0: true graph
  #   adjmat1: estimated graph
  # OUTPUT
  #   fdr: False Discovery Rate
  # ---------------------------------------------------------------
  r0 <- nrow(adjmat0)
  r1 <- nrow(adjmat1)
  c0 <- ncol(adjmat0)
  c1 <- ncol(adjmat1)
  if(r0!=r1 || c0!=c1){
    stop("inconsistent dimensions")
  }
  if(r0!=c0 || r1!=c1){
    stop("not acceptable adjacency matrices")
  }
  TNGraph <- (1-adjmat0)*(1-adjmat1)
  TN      <- sum(TNGraph)
  nonG0   <- adjmat0==0
  FPGraph <- nonG0*adjmat1
  FP      <- sum(FPGraph)
  if(FP+TN==0){
    fpr <- 0
  } else{
    fpr <- FP/(FP+TN)
  }
  return(fpr)
}
FDR <- function(adjmat0, adjmat1){
  # ---------------------------------------------------------------
  # FUNCTION
  #   Compute False Descovery Rate 
  # INPUT
  #   adjmat0: true graph
  #   adjmat1: estimated graph
  # OUTPUT
  #   fdr: False Discovery Rate
  # ---------------------------------------------------------------
  r0 <- nrow(adjmat0)
  r1 <- nrow(adjmat1)
  c0 <- ncol(adjmat0)
  c1 <- ncol(adjmat1)
  if(r0!=r1 || c0!=c1){
    stop("inconsistent dimensions")
  }
  if(r0!=c0 || r1!=c1){
    stop("not acceptable adjacency matrices")
  }
  TPGraph <- adjmat0*adjmat1
  TP      <- sum(TPGraph)
  nonG0   <- adjmat0==0
  FPGraph <- nonG0*adjmat1
  FP      <- sum(FPGraph)
  if(FP+TP==0){
    fdr <- 0
  } else{
    fdr <- FP/(FP+TP)
  }
  return(fdr)
}

power <- function(adjmat0, adjmat1){
  # ---------------------------------------------------------------
  # FUNCTION
  #   Compute power (true positive rate) 
  # INPUT
  #   adjmat0: true graph
  #   adjmat1: estimated graph
  # OUTPUT
  #   pow: power
  # ---------------------------------------------------------------
  r0 <- nrow(adjmat0)
  r1 <- nrow(adjmat1)
  c0 <- ncol(adjmat0)
  c1 <- ncol(adjmat1)
  if(r0!=r1 || c0!=c1){
    stop("inconsistent dimensions")
  }
  if(r0!=c0 || r1!=c1){
    stop("not acceptable adjacency matrices")
  }
  TPGraph <- adjmat0*adjmat1
  TP      <- sum(TPGraph)
  non1    <- (adjmat1==FALSE)*1
  FNGraph <- non1*adjmat0
  FN      <- sum(FNGraph)
  if(FN+TP==0){
    pow   <- 0
  } else{
    pow   <- TP/(FN+TP)
  }
  return(pow)
}

KLdiv <- function(Sigma0, Sigma1, mu0=rep(0,nrow(Sigma0)), mu1=rep(0,nrow(Sigma1))){
  # ---------------------------------------------------------------
  # FUNCTION
  #   Compute KL divergence between two multivariate gaussians
  # INPUT
  #   Sigma0, Sigma1: covariance matrices
  #   mu0, mu1: means
  # OUTPUT
  #   KL: KL(P0||P1)
  # ---------------------------------------------------------------
  
  r0 <- nrow(adjmat0)
  r1 <- nrow(adjmat1)
  c0 <- ncol(adjmat0)
  c1 <- ncol(adjmat1)
  m0 <- length(mu0)
  m1 <- length(mu1)
  if(r0!=r1 || c0!=c1 || m0!=m1 || r0!=m0 || r1!=m1){
    stop("inconsistent dimensions")
  }
  if(r0!=c0 || r1!=c1){
    stop("not acceptable adjacency matrices")
  }
  H <- solve(Sigma1)
  KLmat <- H%*%Sigma0
  KL <- log(abs(det(KLmat)))+tr(KLmat)-d0+(mu1-mu0)%*%H%*%(mu1-mu0)
  return(KL)
}
